#if 0
-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

MODULE: akc_main.c

PROJECT: Autros Keypad Controller firmware

DESCRIPTION:
         Main Module:  Interrupt Service Routine, Initialization,
                       State Machine

AUTHOR:  Kenneth C. Finney (KCF), Wilkes Associates, Inc.

Copyright 1999. All Rights Reserved.

----------------------------------------------------------------------------

REVISIONS & HISTORY:

	KCF 02-08-99 v3110
	   - original code

	KCF 27-08-99 v3111
	   - removed & postponed access code verification functionality,
		  in order to facilitate AMC Auto Config efforts.

	KCF 17-09-99 v3112 -
     -disabled wdt reset code because kpc board needs to
      be modified to support that switch

	KCF 17-09-99 v3113 -
     -removed debug code that toggled charging LED on/off

	KCF 23-09-99 v3114 -
     - re-enabled keypad input code
     - slightly modified access valid buzz
     - moved all of the incoming processing out of the ISR and
       created a finite state machine processor to handle the incoming
       commands.
     - implemented an incoming circular buffer for i2c data

	KCF 23-09-99 v3115 -
     - moved all outgoing i2c processing out of the ISR and into the
	    state machine.

	KCF 23-09-99 v3116 -
     - modified state machine to using fewer states. Changed many states
	    to become commands (state modifiers).

	KCF 06-10-99 v3117 -
     - re-instated the watchdog timer and local reset.
       added remote reset of main controller.
       added diagnostics that indicate current transmit states, by pressing
         the TEST button, certain LEDs are illuminated:

         SERVICE:     WAITING state - waiting for AMC to request a read

         LOW BATTERY: SENDING CODE state - waiting for AMC to finsh
                          reading the access code

         CHARGING:    IDLE state - awaiting keypad presses or AMC commands

         OPEN:        SENDING POP state - waiting for AMC to pick up another
                          POP drawer command

         READY:       (UNUSED) INPUT state - waiting for input at keypad.

        note *** there can be, and often is, more than 1 state active
                  at a given time - most often it is the IDLE state in
                  conjuction with the WAITING state.

	KCF 06-10-99 v3118 -
	  - removed the code that turned off the service LED when CART_ACTIVE_ALERT
	    is received - this is left up to the AMC to explicitely turn off the LED
	    from now on.

	KCF 06-10-99 v3119 -
	  - added support for sleep-mode-cancel by pressing *.

	KCF 06-10-99 v3120 -
	  - modified reboot code.
     
	KCF 06-10-99 v3121 -
	  - fixed bug in reboot code.
-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <pic.h>
#define ULIB_PAUSE_FUNCTIONS
#define ULIB_I2C_FUNCTIONS
#include "akc62a.h"

__IDLOC(3121);  /* controller/firmware revision ID */

/* ---- REVISION ID FORMAT DESCRIPTION -------------------------------------
__IDLOC('a''b''c''d') where
   'a' is Target Board ID -  1 = AMC-01: Autros Main Controller
                             2 = ADC-01: Autros Drawer Controller
                             3 = AKPC: Autros Keypad Controller
   'b' is Target Board Revision:  A,B,C, etc.
   'c' is Firmware Major Version Number
   'd' is Firmware Minor Version Number
------------------------------------------------------------------------- */

/* --------------------- CONFIGURATION WORD FUSES ----------------------- */
__CONFIG(UNPROTECT |  WDTE | XT);
/* no code protect, XT oscillator fuse, powerup timer, watchdog enabled   */


/* ------------------------------ PROTOTYPES ---------------------------- */

void AKC_VerifyLeds(void);
void interrupt AKC_ISR(void);

/* ---------------------------- PUBLIC VARIABLES ------------------------ */

int 	heartbeatCounter = 0;          /* when this is MODed, we change the */
                                      /* heartbeat LED to indicate system */
									                               /* is operating */

BYTE 	i2cData			  = NO_DATA;                 /* I2C, command & state */
bank1 BYTE     	amcCommand       = NO_COMMAND,              /* processing variables */
     	theState         = IDLE_STATE;

BOOLEAN  amcReady2Read = FALSE,   /* */
  		   haveCartCode  = FALSE,
         resetReady = FALSE;

bank1 BYTE  code2Send        = 0;      /* for tracking access code verification */
                                                               /* progess */

BYTE 	nextIn  = 0,            /* these are for use in the inbound & */
				nextGet = 0,            /* outbound I2C circular buffers      */
				nextOut = 0,
				nextPop = 0;

BYTE thePort;

BYTE i2cRcvBuffer[RCV_BUFFER_SIZE];

/* --------------------------- IMPORTED VARIABLES ----------------------- */
extern bank1  BYTE i2cSendBuffer[SEND_BUFFER_SIZE]; /* from akc_kpad.c */
extern struct command_buffer  keypadCommand;        /* from akc_kpad.c */


void AKC_Wait(unsigned int duration) { MilliPause(duration,KICK_DOG);}

/* --------------------------- AKC_ISR ----------------------------------

  The one and only interrupt service routine. Data from master are written
  into the receive circular buffer, from which they are plucked by the state
  machine. When the master wants to read, the ready-to-read flag is set,
  and then the state machine stuffs the transmit buffer.

   ---------------------------------------------------------------------- */
void interrupt AKC_ISR()
{
   i2cData = NO_DATA;
   amcCommand = NO_COMMAND;

   if (SSPIF==TRUE) /* if I2C interrupt occurs */
   {
      SSPIE = FALSE;                    /* disable I2C interrupts for now */
      if (STAT_RW == I2C_MASTER_READ_COMMAND) /* AMC is performing a read */
      {
         amcReady2Read = TRUE;
      }
      else                              /* start I2C_MASTER_WRITE_COMMAND */
      {
         if (STAT_BF == I2C_BUFFER_FULL)
         {
            if (STAT_DA == I2C_DATA_RECEIVED)
            {
               /*                                                */
               /* transfer byte into our circular receive buffer */
               /*                                                */
               i2cRcvBuffer[nextIn++] = SSPBUF;
               if (nextIn >= RCV_BUFFER_SIZE)
                  nextIn=0;
            }      /* end I2C_DATA_RECEIVED */

            else   /* start I2C_ADDRESS_RECEIVED */
            {
               i2cData = SSPBUF; /* empty the buffer to prevent overflow */

            }

         }     /* end I2C_BUFFER_FULL */

      }     /* end I2C_MASTER_WRITE_COMMAND */

      SSPIF = FALSE; /* clear the interrupt notification flag */
      SSPOV = FALSE; /* clear the I2C data buffer overflow flag */
      SSPIE = TRUE;  /* re-enable i2c interrupts */
   }                /* global interrupts are managed by the compiler */
   CLRWDT();

}

/* --------------------------- AKC_Init ----------------------------------

  Initialize tri-state registers, I/O ports and various flags and variables.
  sets up the appropriate options for the PIC chip.

   ---------------------------------------------------------------------- */
void AKC_Init()
{
   RBPU =  ENABLED;        /* PORT B internal weak pullups enabled */
	TRISA = 0x00;           /* keypad row, hbeat LED & buzzer outputs: */
	                        /*  RA0:  keypad button Row 0: '9','0','#' */
	                        /*  RA1:  keypad button Row 1: '4','5','*' */
	                        /*  RA2:  keypad button Row 2: '1','2','3' */
	                        /*  RA3:  keypad button Row 3: '6','7','8' */
	                        /*  RA4:  heartbeat LED                    */
	                        /*  RA5:  buzzer                           */

	TRISC = 0x18;    /* bits 3 & 4 are set to input for I2C slave mode */
	                 /* all others are outputs:                        */

   TRISB = 0xBF; /* keypad column inputs, AMC signalling, test switch */
	                         /*  RB0:  keypad button Column 0: '*','#' */
	                         /*  RB1:  keypad button Column 1: '5','0' */
	                /*  RB2:  keypad button Column 2: '4','9', '3','8' */
	                         /*  RB3:  keypad button Column 3: '2','7' */
	                         /*  RB4:  keypad button Column 4: '1','6' */
	                         /*  RB5:  AMC Reset signal line           */
									 /*        we leave this as an input until */
									 /*        we want to use it by driving it */
									 /*        low after changing to an output */
	                         /*  RB6:  AMC Request signal line         */
	                         /*  RB7:  JP2 'test' switch               */

   PORTA = 0xFF;
   PORTB = 0xDF;
   PORTC = 0x00;

   AKC_Wait(250);
   PORTC = 0xFE;         /* reset LEDs except service led */

	SSPBUF = 0;                         /* clear the I2C data register */

	SSPADD = 0x7E;                           /* fixed I2C base address */
	SSPCON = 0x36;        /* SSP enabled, clock enabled, 7-bit address,*/
	                                              /* no S/P interrupts */
	haveCartCode = FALSE;
   memset(keypadCommand.buffer,0,SEND_BUFFER_SIZE);
   OPTION |= 0x0F;       /* set prescaler to watchdog, set to timeout at ~ 2 sec. */
	PEIE = TRUE;          /* peripheral ints enabled */
	SSPIF = FALSE;        /* clear SSP int. flag */
	SSPIE = TRUE;         /* enable SSP ints */
	GIE = TRUE;           /* global ints enabled */
	SSPOV = FALSE;        /* clear the I2C data buffer overflow flag */
   resetReady = FALSE;
}



/* --------------------------- AKC_Reset ----------------------------------

  Firmware rendition of the PIC reset button. With the watchdog timer enabled,
  when a person presses the 000000 on the AKC board, then either * or # this
  routine is
  invoked, which sets the watchdog prescaler to minimum setting, then enters
  a loop. If the watchdog timer is enabled, it will fire while in this loop,
  thus resetting the chip. If the watchdog is not enabled, then the routine
  will merely spin through the loop, and carry on.

   ---------------------------------------------------------------------- */
void AKC_Reboot()
{
unsigned long i=0xFFFFFFFF;

  mAKC_RebootAMC();                   /* reboot main controller */
  MilliPause(250,KICK_DOG);
  CLRWDT();
  OPTION = 0x0E;       /* set WDT prescaler to timeout at ~ 0.288 sec. */
  while(--i)               /*  wait for WDT to reset akc */
   ;
}



bank1 BYTE outputTarget=0;
bank1 BYTE theOutput=0;

main()
{

   AKC_Init();
   AKC_Blip();

   while(TRUE)
   {
      if (mAKC_CheckState(OUTPUT_STATE))
      {
         mAKC_DoDiagnostic(SLEEP_LED_PIN);
         mAKC_SetState(IDLE_STATE);
         switch(amcCommand)
         {

            case OUTPUT_ON:
                  theOutput=ON;
               break;

            case OUTPUT_OFF:
                  theOutput=OFF;
               break;

            case BEEP:
                  AKC_Beep();
               break;

            case BLIP:
                  AKC_Blip();
               break;

            case CART_ACTIVE_ALERT:
                  AKC_CartActiveAlert();
                  mAKC_SignalAMCReset();
               break;

            case DRAWER_ALREADY_OPEN_ALERT:
                  AKC_DrawerAlreadyOpenAlert();
               break;

            case CODE_IS_VALID:
                  READY_LED_PIN=!READY_LED_PIN;
                  AKC_CodeIsValidBuzz();
                  mAKC_SignalAMCReset();
                  READY_LED_PIN=!READY_LED_PIN;
               break;

            case CODE_NOT_VALID:
                  mAKC_SignalAMCReset();
                  keypadCommand.count = 0;
                  code2Send=0;
		            nextOut=0;
                  haveCartCode=FALSE;
               break;

         } /* end switch-amcCommand */


         switch(outputTarget)
         {
            case SERVICE_LED: SERVICE_LED_PIN=theOutput;break;
            case SLEEP_LED:SLEEP_LED_PIN=theOutput;break;
            case READY_LED:READY_LED_PIN=theOutput;break;
            case LOW_BATT_LED:LOW_BATT_LED_PIN=theOutput;break;
            case CHARGING_LED:CHARGING_LED_PIN=theOutput;break;
            case OPEN_LED:OPEN_LED_PIN=theOutput;break;
            case BUZZER: BUZZER_PIN = theOutput;break;
         }
      } /* end output-state */

      /*                                                                           */
      /* the INPUT STATE is reserved for use if/when we implement interrupt-driven */
      /* keypad scanning & decoding. for now it is unused.                         */

   #if 0
      if (mAKC_CheckState(INPUT_STATE))
      {
		   mAKC_DoDiagnostic(READY_LED_PIN);
      } /* end input-state */
   #endif

      /*                                                                           */
      /* the WAITING STATE is entered when we have signalled the master that we    */
      /* have data to send, but the master has not yet asked us to send it (via    */
      /* an i2c read request).                                                     */
      /*                                                                           */
      if (mAKC_CheckState(WAITING_STATE))
      {
         mAKC_DoDiagnostic(SERVICE_LED_PIN);

         /*                                                                        */
         /* set the state to SENDING_*, but only if we have rec'd a READ commmand  */
         /*                                                                        */
         if(amcReady2Read)
         {
            amcReady2Read = FALSE;
            if(code2Send)               /* transmit a byte of access code */
            {
               mAKC_ChangeState(WAITING_STATE,SENDING_CODE_STATE);
            }
            else
            {                                  /* transmit the drawer pop */
               mAKC_ChangeState(WAITING_STATE,SENDING_POP_STATE);
            }
         }

      } /* end waiting-state */


      /*                                                                           */
      /* the SENDING CODE STATE is when we have access code date queued to send to */
      /* the master, for the next read request it performs. This is for security   */
      /* access codes - pins, cart access, etc.                                    */
      /*                                                                           */
      if (mAKC_CheckState(SENDING_CODE_STATE))
      {
         mAKC_DoDiagnostic(LOW_BATT_LED_PIN);
         if (nextOut < code2Send) /* check if any byte to send, then point to next byte */
         {
            SSPBUF = i2cSendBuffer[nextOut++];   /* send the current byte */
            MicroPause(10);
            mAKC_ReleaseI2C();                         /* release the clock */
         }

         if (nextOut >= code2Send)
         {
            nextOut=0;
            code2Send=0;
            mAKC_ChangeState(SENDING_CODE_STATE,IDLE_STATE);
         }
         else
            mAKC_ChangeState(SENDING_CODE_STATE,WAITING_STATE);
      }

      /*                                                                           */
      /* the SENDING POP STATE is where we want the master to pick up drawer pop   */
      /* commands                                                                  */
      /*
                                                                             */
      if (mAKC_CheckState(SENDING_POP_STATE))
      {
         mAKC_DoDiagnostic(OPEN_LED_PIN);
         mAKC_SignalAMCReset();

         if (nextPop != nextOut)         /* check if any byte to send, */
         {                                  /* then point to next byte */
            i2cData = i2cSendBuffer[nextOut++];/* send the current byte */
				SSPBUF = i2cData;
            if (nextOut >= SEND_BUFFER_SIZE) nextOut = 0;
            MicroPause(10);
            mAKC_ReleaseI2C();                      /* release the clock */
            mAKC_ChangeState(SENDING_POP_STATE,WAITING_STATE);
         }
      }


      /*                                                                       */
      /* the IDLE STATE is our default 'main loop' state, where keypad polling */
      /* is performed, as well as any state transition initiations             */
      /*                                                                       */
      if (mAKC_CheckState(IDLE_STATE))
      {
         mAKC_DoDiagnostic(CHARGING_LED_PIN);
         /*                                                                  */
         /* turn the heartbeat LED on or off, depending on status of counter */
         /*                                                                  */
         if (!(++heartbeatCounter % HEARTBEAT_PULSE) )
            HEARTBEAT_LED_PIN=!HEARTBEAT_LED_PIN;

         /*                                          */
         /* grab the input from keypad, if it exists */
         /*                                          */
         AKC_ProcessKeypad();


         /*                                                            */
         /* clear watchdog now, in case keypad work was time intensive */
         /*                                                            */
         CLRWDT();

         /*                                                   */
         /* grab the next command from the I2C inbound buffer */
         /*                                                   */
         amcCommand = nextGet != nextIn ? i2cRcvBuffer[nextGet++] : NO_COMMAND;
         if (nextGet >= RCV_BUFFER_SIZE) nextGet = 0;


         /*                                               */
         /* extract relevant information from the command */
         /*                                               */
         if ((amcCommand & AKC_CMD_MASK) == OUTPUT_ON)
         {                                    /* turn on LEDs or Buzzer */
            mAKC_SetState(OUTPUT_STATE);
            outputTarget = amcCommand & AKC_OUTPUT_MASK;
            amcCommand = OUTPUT_ON;
         }
         else if((amcCommand & AKC_CMD_MASK) == OUTPUT_OFF)
         {                                    /* turn off LEDs or Buzzer */
            mAKC_SetState(OUTPUT_STATE);
            outputTarget = amcCommand & AKC_OUTPUT_MASK;
            amcCommand = OUTPUT_OFF;
         }
         else if (amcCommand != NO_COMMAND) /* specific alert commands */
         {
            mAKC_SetState(OUTPUT_STATE);
            /* there are no specific output modifiers for these commands */
         }

      } /* end idle-state */

   } /* end while loop */
}

