/*
   AMCMAIN.C   AUTROS MAIN CONTROLLER firmware for MICROCHIP PIC 16F877
*/

//#include </ht-pic/include/stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <pic.h>

/* -------------------------------------------------------------------------
KCF 01-09-99 v1131 -
     added #ifndef TDE to remove code in order to fit it
     in memory for the TDE environment. TDE must be defined in the
     command line options in the TDE project setup in order to
     do the exclusion as follows:  -DTDE

     Added AutoConfig() to config.c

KCF 02-09-99 v1132 -
     removed smart/dumb drawer range checking

KCF 02-09-99 v1133 -
     added SaveConfig to AutoConfig. Rearranged memory bank allocations
     for structures in bank0 & 1 - scalars & arrays to banks 2 & 3

KCF 07-09-99 v1134 -
     recovered 'golden' version after build environment went insane.
     number of drawers reduced to 4 in drawer array.

KCF 15-09-99 v1235 -
     - added conditional compile code for inclusion/exclusion of:
        REDUCE_STACK_USAGE - stack level reduction (by one, in BusyLoop)
        USE_DIAGNOSTICS - diagnostics code
        USE_SERIAL_ECHO - serial echo code
      >> put these in AMC.h and use to juggle memory usages

     - removed the cart/drawer config data save/load code, as well
       as the user entry code for this stuff

KCF 17-09-99 v1236 -
     - revised akc command codes to match current akc rev. (3111)

KCF 21-09-99 v1237 -
     - re-enabled diagnostics code

KCF 23-09-99 v1238 -
     - added backspace support for serial input
     - moved AutoConfig() & Diagnostics() to occur before
       sending cart active alert (instead of after);
     - enabled serial echo diagnostics code.

KCF 13-10-99 v1239 -
     - added hardware hand-shaking (RTS) one-way, for incoming only
     - added serial client bridge monitoring and control code
     - removed the 'idle beep' code and event, and use the available
       event for the scb timeout & reconnect functionality
     - on-board diagnostics are enabled

KCF 21-10-99 v1240 -
     - power/battery management code added

KCF 2-11-99 v1241 -
     - added code to eliminate drawer microswitch 'chatter' (spurious
        unexpected drawer open alerts)
     - diagnostics disabled

KCF 2-11-99 v1242 -
     - revisions to reduce memory usage
     - added stripped down diagnostics back in
     - added 'v' command to view port values and other important variables
       (see flags.txt)
     - added 'x' command to run diagnostics remotely (via serial/telnet)

KCF 2-11-99 v1243 -
     - removed USE_SERIAL_ECHO conditional compile code - serial echo is no
       longer possible (nor needed)
     - added scbAlwaysOn flag to prevent serial client bridge from being
       powered down at sleep time. This mode is used by technicians while
       servicing the SCB (changing settings for example) - they don't want
       the SCB shutting off in the middle of reconfiguring it.
       This mode is invoked by holding down switch S1 (SELECT) during boot
       *before* the cart active alert sounds. If you hold the button down
       beyond the alert, the cart will perform the diagnostic routine.
     - restored 'Z' command for restoring factory configuration defaults

KCF 27-11-99 v1244 -
     - added 1 sec  delay at start-up to allow for drawer initialization
       before autoconfig is executed.

KCF 07-12-99 v1245 -
     - fixed bug in rtc date & time setting - you needed to enter both
       date & time when making a change, not just one or the other, other
       wise the one not changed would be saved as random values
       
KCF 14-12-99 v1246 -
     - changed config display to use lower case labels for differentiation
       from drawer configuration display (which uses upper-case D's )
--------------------------------------------------------------------------- */


__IDLOC(1246);  /* controller/firmware revision ID */

/* ---------- NOTE: controller/firmware revision ID -----------------------

this macro will embed our firmware revision in the ID register of the PIC.
Each digit will acquire a leading0 to pad out the entire value to 8 bytes.
(This is the way that register works) ie. 1234 becomes 01020304

   ---- ID format ----

  __IDLOC('a''b''c''d') where
   'a' is Target Board ID -  1 = AMC-01: Autros Main Controller
                             2 = ADC-01: Autros Drawer Controller
                             3 = AKPC: Autros Keypad Controller
   'b' is Target Board Revision:  A,B,C, etc.
   'c' is Firmware Major Version Number
   'd' is Firmware Minor Version Number
   ------------------------------------------------------------------------ */


/* ------------------------------------------------------------------------
  The following conditional defines are there because early versions of the
  HT-PIC compiler that supported the PIC16F877 was missing them. Leaving them
  in is safest, and harmless at worst.
   ------------------------------------------------------------------------ */
#ifndef LVP
#define LVP                   0x80
#define CPD                   0x100
#define WRT                   0x200
#define BKBUG                 0x800
#define HS                    FOSC1
#define RC              (FOSC1 | FOSC0)
#define XT              FOSC0
#define LP              0x00
#endif


__CONFIG(UNPROTECT | WRT | CPD | BKBUG | HS);
/* ----------------- NOTE: Configuration Fuses ----------------------------
   The above fuses are set as follows:

   no code protect, powerup timer enabled, low voltage program enabled,
   flash program enabled, data ee protect disabled, HS oscillator fuse
   in circuit debugging disabled
   ------------------------------------------------------------------------ */


#include "amc.h"          /* general amc definitions & data structures */
#include "config.h"       /* auto-config, config & display */
#include "i2c.h"          /* app-level i2c stuff */
#include "drawers.h"
#include "timer.h"
#include "remoteio.h"
#include "power.h"

/* command declarations */


struct command_buffer  command;   /* buffer for RS-232 commands */
BYTE                   current_drawer;
BYTE                   bin;
bank1 BYTE                   scbSleepTimeout;
bank1 struct bin_mask        bin_status;

struct cart_configuration cart;
bank1 struct memory_location location = {0x0,0x0};
bank2 BYTE activeDrawerMask = 0x00;
bank2 unsigned int heartbeatCounter = 0;

#if USE_RING_BUFFER
bank2 BYTE ringBuffer[MAX_RING_BUFFER_LENGTH], nextIn=0;
#endif

bank2 BYTE kpCommand=0;
BYTE scbStatus=SCB_OFFLINE;
bank2 BYTE volts=0;


/* ------------------------------------------------------------------------
   These bytes are used in power.c to monitor and regulate the power mgmt
   system.
   ------------------------------------------------------------------------ */
BYTE  batteryState=BATTERY_DEAD,
      powerState=POWER_BAD,
      lastBatteryState=BATTERY_DEAD,
      lastPowerState=POWER_BAD;

/* ------------------------------------------------------------------------
   The following flags monitor various internal states of the firmware.
   Their values can be viewed at runtime by using the 'V' command at the
   host interface
   ------------------------------------------------------------------------ */
BOOLEAN  busy_looping,
         aDrawerIsOpen,
         idle,
         valid_cart_code,
         openExpected,
         kpcWasHandled,
         cartIsActive,
         scbAlwaysOn,
         doPowerUpdate,
         forceSleep,
         sleepLEDstate;



/* ------------------------------------------------------------------------

   Function:      AMC_Test

   Parameters:    BYTE direct   - If TRUE then can be invoked from
                                  from command interface (or elsewhere in
                                  the code).

                                - FALSE means it must be invoked by
                                  pressing the SELECT button on the AMC
                                  main controller. To make this funciton
                                  execute indirectly, you must call it
                                  periodically in a loop - it will then
                                  check the state of the SELECT switch.

   Returns:       none

   Description: Performs a quick test routine to help troubleshoot
                problems in the field.

                HOW TO USE TEST FUNCTION

                  Diagnostics are invoked/enabled by pressing SW1 on the AMC.
                  The following diagnostic functions will be performed when
                  SW1 is pressed while the cart is active (ie. READY LED is
                  lit on keypad):

                1) keypad test. will light all keypad LEDS in order from
                   SERVICE LED to LOW BATTERY LED in an anti-clockwise
                   direction.

                2) drawer test. All drawers will be popped, in order. Smart
                   drawers will pop, and then cycle through lighting their
                   bin LEDS. (a future version will test the reed switches
                   also).

   ------------------------------------------------------------------------ */
#if USE_TEST
void AMC_Test(BYTE direct)
{
static bank3 BYTE d,b;


   if (RE1==OFF || direct) /* switch 'Select' S1 or directly commanded from code*/
   {
     idle=FALSE;

    // diagnosticSwitches = (PORTB & DIAGNOSTIC_SWITCH_MASK) >> 4;  NOT USED

    for (d=1;d<9 ;d++)
    {
      AMC_KpLed(d, LED_ON);
      if (drawer[d-1].i2c_address)
      {
        AMC_PopDrawer(d);
        MilliPause(54);
        for (b=1;b<=24 ;b++)
        {
          AMC_WriteDrawer(d,b);
        }
      }
      else
        AMC_PopDrawer(d);
      MilliPause(100);
      AMC_KpLed(d, LED_OFF);
     } /* end-for */
     AMC_KpLed(READY_LED, LED_ON);
   } /* end-if*/
}

#endif

/* ------------------------------------------------------------------------

   Function:      AMC_GetKPCommand

   Parameters:    none

   Returns:       BYTE  - the embedded bytecode command extracted from the I2C
                          read.


   Description:   Executes an i2c READ command to the KPC keypad controller.
                  if the bytestream was a command, then it is extracted and
                  returned. If the bytestream was an access code, then
                  command.buffer is populated with the access code values
                  for later verification.

   ------------------------------------------------------------------------ */
BYTE AMC_GetKPCommand()
{
BYTE byteCtr;
signed char i2cResult;

  byteCtr=0;
  U_I2CStart();
  i2cResult = U_I2CWrite(KEYPAD_I2C_ADDRESS+1);
  AMC_BusyLoop(I2C_READ_BYTE_DELAY);
  i2cResult = U_I2CRead(I2C_LAST);
  command.count=0;
  if (i2cResult==REQUEST_CODE)
  {
     AMC_BusyLoop(I2C_READ_BYTE_DELAY);
     U_I2CStart();
     i2cResult = U_I2CWrite(KEYPAD_I2C_ADDRESS+1);
     AMC_BusyLoop(I2C_READ_BYTE_DELAY);
     command.count = U_I2CRead(I2C_MORE);
     AMC_BusyLoop(I2C_READ_BYTE_DELAY);
     for(byteCtr=0;byteCtr<(command.count - 1);byteCtr++)
     {
       command.buffer[byteCtr] = U_I2CRead(I2C_MORE);
       AMC_BusyLoop(I2C_READ_BYTE_DELAY);
     }
     command.buffer[byteCtr] = U_I2CRead(I2C_LAST);
  }
  AMC_BusyLoop(I2C_READ_BYTE_DELAY);
  U_I2CStop();
  return ((BYTE)i2cResult);
}



/* ------------------------------------------------------------------------

   Function:      AMC_ExecuteKPCommand

   Parameters:    BYTE theCommand   - command bytecode extracted from i2c
                                      read by AMC_GetKPCommand

   Returns:       BYTE              - TRUE in all cases, except if the
                                      command was validate an access code
                                      in which case, we return TRUE if
                                      valid, FALSE if not valid

   Description:   Extracts the opcode from the bytecode, and any embedded
                  parameters, then executes based on command. Valid executives
                  are:
                     KP_POP_DRAWER - pop open the drawer whose ID is embedded.

                     validate access code, whose byte stream has already been
                     captured in command.buffer

   ------------------------------------------------------------------------ */
BYTE
AMC_ExecuteKPCommand(BYTE theCommand)
{
static bank3 BYTE value=0;

  if ( ((theCommand & KP_DRAWER_CMD_MASK) +KP_SAFETY_NIBBLE) == KP_POP_DRAWER)
  {
    if (valid_cart_code==TRUE)
    {
       openExpected=TRUE;
       aDrawerIsOpen=FALSE;
       value = theCommand & KP_DRAWER_MASK;
       if  ( value && (activeDrawerMask & (0x01 << (value-1))) )
         AMC_PopDrawer(value);
    }
  }
  else if (command.count)  /* it's an access code request */
  {
     valid_cart_code = FALSE;
    for(value=0;value<MAX_ACCESS_CODE_LENGTH;value++)
    {
      if (command.buffer[value] != cart.accessCode[value])
      {
        AMC_KPAccessResult(CODE_NOT_VALID );
        return 0;
      }
     }
      valid_cart_code = TRUE;
     AMC_KPAccessResult(CODE_IS_VALID);
  }

  return (1);
}



/* ------------------------------------------------------------------------

   Function:      AMC_ProcessCommand

   Parameters:    none

   Returns:       none

   Description:   executes the command parsed by parse_command.

   ------------------------------------------------------------------------ */

void AMC_ProcessCommand()
{
  if (command.count > 2 )
    current_drawer = atoi(&command.buffer[2]);
  switch ( command.buffer[1] )
  {
#if USE_RING_BUFFER
    case 'B':
       {
      BYTE i;
      putch('<');
      for (i=0;i<MAX_RING_BUFFER_LENGTH;i++)
      putch(ringBuffer[i]);
      putch('>');
      break;
    }
#endif

    case 'C' :
       if (command.buffer[0]!='?')
       {
          mAMC_FlowinOn();
          AMC_Config();             /* fall thru to next line if to display */
       }
       AMC_DisplayConfig();
    break;

    case 'D' :
       openExpected=TRUE;
       aDrawerIsOpen=FALSE;
       if (!drawer[current_drawer-1].i2c_address) /* if no i2c address,then */
         AMC_PopDrawer(current_drawer); /* then I'm a matrix(dumb) drawer ! */
    break;

    case 'S' :
      if (command.count < 5)
        break;
      bin = atoi(&command.buffer[4]);
      if ( bin < 1 || bin > 24)
        break;
      openExpected=TRUE;
      aDrawerIsOpen=FALSE;
      command.buffer[3] = '\0';
      if (drawer[current_drawer-1].i2c_address)      /* if have i2c address */
      {                                              /* then I ain't dumb ! */
        AMC_WriteDrawer(current_drawer,bin);
        AMC_PopDrawer(current_drawer);
      }
      break;

#if USE_DIAGNOSTICS
    case 'T' :                     /* temporary command to test smart drawer */
      if (drawer[current_drawer-1].i2c_address)/* =i2c addr. ? smart drwr   */
      {
        AMC_ReadDrawer(current_drawer,&bin_status);
        U_PrintHex(bin_status.bin1to8);
        U_PrintHex(bin_status.bin9to16);
        U_PrintHex(bin_status.bin17to24);
      }
    break;

    case 'V' :
       putch('\n');putch('\r');
       U_PrintDec(batteryState);
       U_PrintDec(powerState);
       U_PrintDec(scbStatus);
       U_PrintDec(volts);
       U_PrintHex(PORTA);
       U_PrintHex(PORTB);
       U_PrintHex(PORTC);
       U_PrintHex(PORTD);
       U_PrintHex(PORTE);

       putch('%');
       U_PrintFlag(busy_looping);
       U_PrintFlag(aDrawerIsOpen);
       U_PrintFlag(idle);
       U_PrintFlag(valid_cart_code);
       U_PrintFlag(openExpected);
       U_PrintFlag(kpcWasHandled);
       U_PrintFlag(cartIsActive);
       U_PrintFlag(scbAlwaysOn);
       U_PrintFlag(doPowerUpdate);
       U_PrintFlag(forceSleep);
       U_PrintFlag(sleepLEDstate);
    break;

#endif

#if USE_TEST
    case 'X' :
       AMC_Test(1);
    break;
#endif

    case 'Z' :
      mAMC_SETDEFAULTCONFIGURATION();
     AMC_SaveConfig();
    break;

   default:
     break;

  }
  command.count = 0;

}



/* ------------------------------------------------------------------------

   Function:   AMC_ParseCommand

   Parameters:

   Returns:

   Description:
               Processes a stream of characters from the rs-232c serial input
               and parses it into commands

   ------------------------------------------------------------------------ */
BYTE AMC_ParseCommand(char c)
{
  if (c == 0x1b)  /* use escape to cancel command */
  {
         command.count=0;
//         memset(&command,0,sizeof(struct command_buffer));
         return (0);
  }
  if (c == 0x08)  /* backspace */
  {
         if(command.count)
          --command.count;
         return (0);
  }

  if ( command.count == 0 )
  {
    if (c=='?')
    {
       command.buffer[command.count++] = c;
       return (0);
    }
    else
    {
       c = toupper(c);
       if ( (c >= 'A') && (c <= 'Z') )
       {
          command.buffer[command.count++] = ' ';
          command.buffer[command.count++] = c;
       }
       return (0);
    }
  }

  if ( command.count == 1 )
  {
    c = toupper(c);
    if ( (c >= 'A') || (c <= 'Z') )
      command.buffer[command.count++] = c;
    return (0);
  }

  if (c != ';')  // command not complete
  {
    c = toupper(c);
    if ( (c == ',') || ((c >= '0') && (c <= '9')) || ((c >= 'A') && (c <= 'Z')))
        command.buffer[command.count++] = c;
    return (0);
  }
  command.buffer[command.count] = '\0';
  return (1);
}



/* ------------------------------------------------------------------------

   Function:      AMC_ProcessSerial

   Parameters:

   Returns:

   Description: reads and processes any characters from the rs-232 port.

   ------------------------------------------------------------------------ */
void AMC_ProcessSerial()
{
static bank3 char charIn;

   while(mU_KBHIT())    /* checks for serial character in shift register */
   {
     mAMC_FlowinOff();
     charIn=getch();   /* fetch the character */
#if USE_RING_BUFFER

     ringBuffer[nextIn++] = charIn;
     if (nextIn >= MAX_RING_BUFFER_LENGTH) nextIn = 0;
#endif
     U_SerialCheckOERR(); /* clean up the over-run bit, if set */
     if ( AMC_ParseCommand(charIn) )
       AMC_ProcessCommand();
     mAMC_FlowinOn();
     idle = FALSE;
   }
}


/* ------------------------------------------------------------------------

   Function:      AMC_ProcessDrawersOpen

   Parameters:

   Returns:

   Description:
         checks for open drawers. If any are detected the function creates
         a timer event which causes the buzzer to sound after the
         timeout. If the drawer is closed before the timeout, then the
         reset_drawers_open function will remove the timer event.

   ------------------------------------------------------------------------ */
void AMC_ProcessDrawersOpen()
{
  if (!openExpected)
  {
    AMC_BusyLoop(250); /* wait a bit to make sure that this is not microswitch 'chatter' */
    if (AMC_DrawersOpen() & activeDrawerMask) /* check again ... */
    {
      AMC_SendAlert(DRAWER_ALREADY_OPEN_ALERT);
      AMC_BusyLoop(750);
      AMC_ResetFastEvent(DRAWER_OPEN_EVENT,cart.drawerOpenTimeout);
      openExpected=TRUE;
      aDrawerIsOpen=TRUE;
      AMC_KpLed(OPEN_LED, LED_ON);
    }
    else
      return;
  }

  if (!aDrawerIsOpen)
  {
    AMC_KpLed(OPEN_LED, LED_ON);
     aDrawerIsOpen = TRUE;
    AMC_ResetLongEvent(DRAWER_OPEN_EVENT,cart.drawerOpenTimeout);
  }

  /* else the timer event was already set up */
  if ( current_drawer && drawer[current_drawer-1].i2c_address ) // check for wrong bin open
  {
    if ( bin == 0 )
    {
      AMC_KPBeep(BUZZER_OFF);
      return;
    }
    AMC_ReadDrawer(current_drawer,&bin_status);
    AMC_WriteDrawer(current_drawer,bin);

    if (( bin < 9 ) && (bin >= 1))
        bin_status.bin1to8 &= (~(1 << (bin - 1)));
    else if (( bin < 17 ) && (bin > 8))
        bin_status.bin9to16 &= (~(1 << (bin - 9)));
    else if ( (bin < 24 ) && (bin > 16))
        bin_status.bin17to24 &= (~(1 << (bin - 17)));

    if (bin_status.bin1to8 || bin_status.bin9to16 || bin_status.bin17to24 )
    {
      AMC_KPBeep(BUZZER_ON);
    }
    else
    {
       AMC_KPBeep(BUZZER_OFF);
    }
  }
}


/* ------------------------------------------------------------------------

   Function:      AMC_ResetDrawersOpen

   Parameters:

   Returns:

   Description:
            removes the timer event when all drawers are shut and turns off
           the open LED and buzzer (in case they were already on).
   ------------------------------------------------------------------------ */
void AMC_ResetDrawersOpen()
{
 if (aDrawerIsOpen)
 {
   AMC_KPBeep(BUZZER_OFF);
   AMC_KpLed(OPEN_LED, LED_OFF);
   aDrawerIsOpen = FALSE;
   current_drawer = 0;
   bin = 0;
   AMC_RemoveTimerEvent(DRAWER_OPEN_EVENT);
   openExpected=FALSE;
  }
}




/* ------------------------------------------------------------------------

   Function:      AMC_InitPorts

   Parameters:

   Returns:

   Description:  initialize the port values, set direction registers

   ------------------------------------------------------------------------ */
void
AMC_InitPorts()
{
   ADCON0 = 0x80; /* select Fosc/32 for A/D conversion */

/* ------------------------------------------------------------------------
   The ClearView Mathias ICE does not support the 16F877 yet, so we
   use the module for the 16C77: therefore we have conditionally compile
   certain features such as the A/D port setup which follows. The
   configurations are different, plus the C77 has fewer ports.
   ------------------------------------------------------------------------ */
#if defined (_16F877)
   ADCON1 = 0x7E; /* disable A/D on Ports A & E, except RA0 which is analog */
#else
   ADCON1 = 0x05; /* disable A/D on Ports A & E, except RA0 which is analog */
                  /* when emulating in C77 mode, RA1 is also analog !! */
#endif


   RBPU = 0;                      /* port B weak pullups enabled (0=enabled)*/
   /*using internal pullups helps keep the part population low on the board */



/* ------------------------------------------------------------------------
   See Bill Richardson's application notes, or the Tech Ref manual for an
   explanation of the direction registers and ports whose setup is below
   ------------------------------------------------------------------------ */
   TRISA = 0x3B;
   TRISB = 0xFF;
   TRISC = 0x99;
   TRISD = 0xF0;
   TRISE = 0x03;

   PORTA = 0x00;
   PORTB = 0x00;
   PORTC = 0x20;
   PORTD = 0x00;
   PORTE = 0x00;
}




/* ------------------------------------------------------------------------

   Function:      main

   Parameters:    none

   Returns:       none

   Description:

   Implementation of main procedure. Most peripherals are controlled from
   routines in other modules. Heartbeat is managed from the main loop.
   Main loop is entered after boot process is complete. AKC is notified that
   cart is active when the main loop is entered.

   ------------------------------------------------------------------------ */
void main()
{
/* ------------------------------------------------------------------------
   This delay is inserted to ensure that the ADC drawer controllers are fully
   booted before we query them with AutoConfig()
   ------------------------------------------------------------------------ */
  MilliPause(1000);

  cartIsActive=FALSE;                 /* stays like this until fully booted */
  scbAlwaysOn = FALSE; /*unless changed below, scb will sleep when timedout */

  AMC_InitPorts();   /* initialize the port values, set direction registers */



/* ------------------------------------------------------------------------
If the select switch is held down at boot time, for the first second or so,
scbAlwaysOn will be set to TRUE. Then the scb will NOT sleep after timeout
   ------------------------------------------------------------------------ */
  if (!RE1)         /* check if S1 (SELECT) is being pressed */
  {
    MilliPause(30);  /* crude debounce wait */
    if (!RE1)
    {
      AMC_KPBeep(BEEP);    /* indicate mode has been set */
      scbAlwaysOn = TRUE;  /* if S1 pressed, set to always on (no sleep) */
    }
  }


  AMC_OffScb();
  AMC_KpLed(SLEEP_LED, LED_ON);
  U_SerialInit(BPS);
  AMC_KpLed(SERVICE_LED, LED_ON);
  AMC_InitTimer();
  mU_ENABLE_SERIAL_PORT(); /* 'cos we're gonna need it */
  AMC_InitI2C();
  scbSleepTimeout = cart.sleepTimeout;
  AMC_UpdatePower();                 /* check power status & install event */
  mU_DISABLE_SERIAL_PORT();  /* finished with config */
                             /* so allow AMC_Wake to do the job */

  current_drawer = bin = command.count = command.count = 0;
  valid_cart_code=FALSE;

  AMC_AutoConfig();
  AMC_Wake();

  openExpected=FALSE;

  kpcWasHandled = FALSE;
  HEARTBEAT_LED_PIN=OFF;

  doPowerUpdate=TRUE;
  cartIsActive = TRUE;
  MilliPause(200);
  AMC_SendAlert(CART_ACTIVE_ALERT); /* A/V indicater that cart is active now */

  MilliPause(600);

  while(TRUE)
  {
     if (doPowerUpdate)
       AMC_UpdatePower();              /* check power status & install event */

#if USE_DIAGNOSTICS
     AMC_Test(0);
#endif

     if (!(++heartbeatCounter%HEARTBEAT_PULSE_TIME))
     {
          mAMC_TOGGLEHEARTBEATLED();
     }


    if ( (KPC_REQUEST==NOT_ASSERTED) && (kpcWasHandled==TRUE))
    {                              /*if keypad does not want our attention */
       kpcWasHandled = FALSE;
    }
    else if ( (KPC_REQUEST==ASSERTED) && (kpcWasHandled==FALSE) )
    {                                     /* if keypad wants our attention */
        idle = FALSE;
        if (batteryState)
           forceSleep=FALSE;
        MilliPause(1);        /* delay to allow kpc to set state variables */
        kpCommand=AMC_GetKPCommand();
        AMC_ExecuteKPCommand(kpCommand);
        kpcWasHandled = TRUE;
     }

    AMC_ProcessSerial();
    AMC_UpdateTime();
    AMC_ProcessEvents();

    if (AMC_DrawersOpen() & activeDrawerMask)
      AMC_ProcessDrawersOpen();
    else
      AMC_ResetDrawersOpen();

    if ( bin && current_drawer && drawer[current_drawer-1].i2c_address )
      AMC_WriteDrawer(current_drawer,bin);      /* check for wrong bin open */

  if (scbStatus==SCB_ONLINE)
  {
    if (!idle)
    {
      AMC_ResetLongEvent(SLEEP_EVENT,scbSleepTimeout);
      idle=TRUE;
    }

    if ( (!AMC_DCD || forceSleep)  )            /* if connection is lost... */
    {                           /* or power manager decides to shut SCB off */
      scbStatus = SCB_OFFLINE;
      AMC_OffScb();
      AMC_KpLed(SERVICE_LED, LED_ON);
      AMC_ResetFastEvent
          (SCB_RECONNECT_EVENT,cart.scbConnectTimeout *
                          SCB_RECONNECT_FACTOR * 1000);
      idle=TRUE;
    }

  }
  else
  {
    if (!idle)/* if no longer idle, wake-up scb if it's off */
    {
      if (!forceSleep)
      {
        AMC_Wake();
      }
    }
   }


  }  /* end of main while-loop */
}

