// timer.h
// These routines initialize and start counters and maintain a list of
// timed events. Each event specifies a callback routine which will be called once
// at the specified time. To have a routine called repeatedly at a specified interval,
// just have the callback routine reset the timer event. The times are expressed in
// minutes, and milliseconds.
///////////////////////////////////////////////////////////////////////////////////////

#include <pic.h>
#include <string.h>


#include "amc.h"
#include "timer.h"
#include "remoteio.h"
#include "power.h"

extern BOOLEAN   doPowerUpdate;
extern BOOLEAN   scbAlwaysOn;
extern BYTE      batteryState;
extern bank1 BYTE scbSleepTimeout;

extern BYTE      scbStatus;

extern BOOLEAN    busy_looping,
                  aDrawerIsOpen,
                  idle;

bank2 long        minute_count;// @ 0x110;  // A running minutes counter
bank2 long        msecond_count;// @ 0x114; // A running millisecond counter

bank1 struct  timer_event      event[NUM_EVENTS];


//////////////////////// off_scb ///////////////////////////
// turns off the serial client bridge when power is low or
// the cart has been idle for the timeout period
///////////////////////////////////////////////////////////
void AMC_OffScb()
{
  if (scbAlwaysOn)
    return;
  mAMC_DISABLE_SCB;
  /* avoid cycling power to scb too fast for it to recover */
 // MilliPause(2000);
  AMC_UpdateTime();
}


//////////////////////// on_scb ////////////////////////////////
// turns on the serial client bridge after activity is detected
////////////////////////////////////////////////////////////////
void AMC_OnScb()
{
 // AMC_KpLed(LOW_BATT_LED, LED_OFF);
  scbStatus = SCB_ONLINE;
  mAMC_ENABLE_SCB;
}

/////////////////////// wake //////////////////////////////
// turns on the serial client bridge, waits n seconds and
// turns off sleep light
///////////////////////////////////////////////////////////
void AMC_Wake()
{
static bank3 unsigned int timeout;

   if (batteryState == BATTERY_DEAD)
   {
	  mU_ENABLE_SERIAL_PORT();
     return;
   }
    mAMC_FlowinOff();
    mU_DISABLE_SERIAL_PORT(); /* because the scb can't boot otherwise */

    AMC_KpLed(READY_LED, LED_OFF);
    AMC_KpLed(SERVICE_LED, LED_ON);
    AMC_UpdateTime();

    AMC_OnScb();

    timeout = cart.scbConnectTimeout*10;
    while (!AMC_DCD && --timeout)
      MilliPause(50);                // wait until scb is rebooted

	  mU_ENABLE_SERIAL_PORT();

    AMC_KpLed(SLEEP_LED, LED_OFF);
    AMC_KpLed(READY_LED, LED_ON);
    if (AMC_DCD)
	 {
      scbStatus=SCB_ONLINE;
      AMC_KpLed(SERVICE_LED, LED_OFF);
      AMC_ResetLongEvent(SLEEP_EVENT,scbSleepTimeout);
	 }
	 else
	 {
      scbStatus=SCB_OFFLINE;
      AMC_ResetFastEvent(SCB_RECONNECT_EVENT,cart.scbConnectTimeout * SCB_RECONNECT_FACTOR * 1000);
	 }
	 mAMC_FlowinOn();
    idle = TRUE;
}

//////////////////// busy /////////////////////////////////
// event to tell busy loop when it is done
///////////////////////////////////////////////////////////
void AMC_Busy()
{
  busy_looping = FALSE;
}

/////////////////////////////// update_time /////////////////////////////////////////
// Purpose: to keep a running count of time
//////////////////////////////////////////////////////////////////////////////////////
void AMC_UpdateTime()
{
unsigned int timer_value;
unsigned int limit;

  limit=59999L;
  timer_value = TMR1L;
  if ( timer_value < 4 )
    return;
  TMR1L= (timer_value % 4);
  if ((msecond_count += (timer_value >> 2)) > limit)
  {
    msecond_count = 0L;
    ++minute_count;
  }
}

/////////////////////////////// process_events ///////////////////////////////////////
// Purpose: to process any scheduled events
//////////////////////////////////////////////////////////////////////////////////////
void AMC_ProcessEvents()
{
//  mAMC_ProcessEvents;
int current;int past_time;

  for (current = 0; current < NUM_EVENTS; current++)
  {
    if (event[current].milliseconds )
    {
      past_time = FALSE;
      if (minute_count > event[current].minutes)
        past_time = TRUE;
      else if (minute_count == event[current].minutes)
        if ( msecond_count >= event[current].milliseconds)
          past_time = TRUE;

      if (past_time)
      {
        event[current].milliseconds = 0;
        switch (current)
        {
          case DRAWER_OPEN_EVENT:
			 	AMC_KPBeep(BEEP);
				MilliPause(30);
			 	AMC_KPBeep(BEEP);
				AMC_ResetFastEvent(current, 5000);
				break;
          case POWER_TIMEOUT_EVENT:
				 	doPowerUpdate=TRUE;
			 	break;
          case BUSY_EVENT:
			 	AMC_Busy();
				break;
          case SLEEP_EVENT:
            if (scbStatus==SCB_OFFLINE)
              break;
			 	AMC_OffScb();
            AMC_KpLed(SLEEP_LED, LED_ON); /* do this here to avoid stack overflow */
            scbStatus = SCB_SLEEPING;
           // idle = TRUE; /* to keep from waking up scb immediately after putting it to sleep */
				break;
          case SCB_RECONNECT_EVENT:
            if (scbStatus==SCB_SLEEPING)
              break;
			 	AMC_OffScb();
            scbStatus = SCB_OFFLINE;
		      idle=FALSE;
				break;
          default:    break;
        }/* end-switch */
      } /* end-if past_time */

    }
  }
}

void AMC_ResetLongEvent(int event_index, long minutes)
{
  if (minutes==0)
    return;
  if (event_index < NUM_EVENTS)
  {
    event[event_index].minutes = minute_count + minutes;
    event[event_index].milliseconds = msecond_count?msecond_count:1;
   }
}

void AMC_ResetFastEvent(int event_index, long milliseconds)
{
long ltemp;
  if (milliseconds==0)
    return;

  if (event_index < NUM_EVENTS)
  {
    if ( (59999L - msecond_count) < milliseconds )
    {
      event[event_index].minutes = minute_count + 1;
      ltemp = 60000L - milliseconds;
      event[event_index].milliseconds = msecond_count - ltemp;
    }
    else
    {
      event[event_index].minutes = minute_count;
      event[event_index].milliseconds = msecond_count + milliseconds;
    }
    if ( !event[event_index].milliseconds )
      event[event_index].milliseconds = 1;
  }
}

void AMC_RemoveTimerEvent(int event_index)
{
  if (event_index < NUM_EVENTS)
    if ( event[event_index].milliseconds )
        event[event_index].milliseconds = 0;
}

/////////////////////// busy_loop //////////////////////////
// a replacement for delay_ms() that doesn't mess up the
// timers.
///////////////////////////////////////////////////////////
void AMC_BusyLoop(long milliseconds)
{
  busy_looping = TRUE;
  AMC_ResetFastEvent(BUSY_EVENT,milliseconds);
  while ( busy_looping )
  {
    AMC_UpdateTime();
    AMC_ProcessEvents();  /* use this when using TDE, to save memory */
  }
}


void AMC_InitTimer()
{
int i;

  busy_looping   = FALSE;
  aDrawerIsOpen  = TRUE;
  idle           = FALSE;
  TMR1L = 0;
  TMR1H = 0;
  T1CON=0x03;

  OPTION = 0x8F;    /* OPTION bits described below */
#if 0
	T0CS = TRUE;   /* bit 5  selects external clock source on TOCK1 pin */
	T0SE = FALSE;  /* bit 4  increment timer0 on external timer low-to-high */
	PSA = FALSE;   /* bit 3  Assign prescaler to wdt */
	PS2 = TRUE;    /* bit 2  "" */
	PS1 = TRUE;    /* bit 1  "" */
	PS0 = TRUE;    /* bit 0  these 3 set WDT to 2304MS */
#endif

   for (i=0;i<NUM_EVENTS;i++)
	{
	  event[i].minutes=0;
	  event[i].milliseconds=0;
   }

	PEIE = TRUE;		   // peripheral interrupt enable
	GIE = TRUE;		      // Global interrupt enable
	minute_count = msecond_count = 0L;
}
